from abc import ABCMeta, abstractmethod
from typing import Any

from jellyfish import (
    jaccard_similarity,
    jaro_similarity,
    jaro_winkler_similarity,
    levenshtein_distance,
)
from matchescu.matching.similarity._common import Similarity


class StringSimilarity(Similarity, metaclass=ABCMeta):
    def __init__(self, ignore_case: bool = False):
        self.__ignore_case = ignore_case

    @abstractmethod
    def _compute_string_similarity(self, x: str, y: str) -> float:
        pass

    def _compute_similarity(self, a: Any, b: Any) -> float:
        x = str(a or "")
        y = str(b or "")

        if self.__ignore_case:
            x = x.lower()
            y = y.lower()

        return self._compute_string_similarity(x, y)


class LevenshteinDistance(StringSimilarity):
    def _compute_string_similarity(self, x: str, y: str) -> float:
        return levenshtein_distance(x, y)


class LevenshteinSimilarity(StringSimilarity):
    def _compute_string_similarity(self, x: str, y: str) -> float:
        m = len(x)
        n = len(y)

        if m == 0 and n == 0:
            return 1.0
        if m == 0 or n == 0:
            return 0.0

        relative_distance = levenshtein_distance(x, y) / max(m, n)
        return round(1 - relative_distance, ndigits=2)


class Jaro(StringSimilarity):
    def _compute_string_similarity(self, x: str, y: str) -> float:
        return jaro_similarity(x, y)


class JaroWinkler(StringSimilarity):
    def _compute_string_similarity(self, x: str, y: str) -> float:
        return jaro_winkler_similarity(x, y)


class Jaccard(StringSimilarity):
    def __init__(self, ignore_case: bool = False, threshold: int | None = None):
        super().__init__(ignore_case)
        self.__threshold = threshold

    def _compute_string_similarity(self, x: str, y: str) -> float:
        y_len = len(y)
        x_len = len(x)
        threshold = self.__threshold or min(x_len, y_len)
        if threshold == 0:
            return 0 if x_len > 0 or y_len > 0 else 1

        return jaccard_similarity(x, y, threshold)



import logging
import os
import sys
import time
import warnings
from contextlib import contextmanager
from functools import partial
from pathlib import Path

import polars as pl
from transformers import AutoTokenizer

from matchescu.blocking import TfIdfBlocker
from matchescu.comparison_filtering import is_cross_source_comparison
from matchescu.data_sources import CsvDataSource
from matchescu.extraction import (
    Traits,
    RecordExtraction,
    RecordIdAdapter,
    single_record,
)
from matchescu.csg import BinaryComparisonSpaceGenerator, BinaryComparisonSpaceEvaluator
from matchescu.matching.ml.ditto._ditto_dataset import DittoDataset
from matchescu.matching.ml.ditto._ditto_module import DittoModel
from matchescu.matching.ml.ditto._ditto_trainer import DittoTrainer
from matchescu.matching.ml.ditto._ditto_training_evaluator import DittoTrainingEvaluator
from matchescu.reference_store.id_table import InMemoryIdTable
from matchescu.typing import EntityReferenceIdentifier

DATADIR = os.path.abspath("data")
BERT_MODEL_NAME = "roberta-base"
LEFT_CSV_PATH = os.path.join(DATADIR, "abt-buy", "Abt.csv")
RIGHT_CSV_PATH = os.path.join(DATADIR, "abt-buy", "Buy.csv")
GROUND_TRUTH_PATH = os.path.join(DATADIR, "abt-buy", "abt_buy_perfectMapping.csv")

# set up abt extraction
abt_traits = list(
    Traits().int(["id"]).string(["name", "description"]).currency(["price"])
)
abt = CsvDataSource(LEFT_CSV_PATH, traits=abt_traits).read()
# set up buy extraction
buy_traits = list(
    Traits()
    .int(["id"])
    .string(["name", "description", "manufacturer"])
    .currency(["price"])
)
buy = CsvDataSource(RIGHT_CSV_PATH, traits=buy_traits).read()
# set up ground truth
gt = set(
    (
        EntityReferenceIdentifier(x[0], abt.name),
        EntityReferenceIdentifier(x[1], buy.name),
    )
    for x in pl.read_csv(
        os.path.join(DATADIR, "abt-buy", "abt_buy_perfectMapping.csv"),
        ignore_errors=True,
    ).iter_rows()
)

log = logging.getLogger(__name__)


def create_comparison_space(id_table, ground_truth, initial_size):
    csg = (
        BinaryComparisonSpaceGenerator()
        .add_blocker(TfIdfBlocker(id_table, 0.23))
        .add_filter(is_cross_source_comparison)
    )
    comparison_space = csg()
    eval_cs = BinaryComparisonSpaceEvaluator(ground_truth, initial_size)
    metrics = eval_cs(comparison_space)
    print(metrics)
    return comparison_space


def _id(record, source):
    return EntityReferenceIdentifier(record[0], source)


def load_data_source(id_table: InMemoryIdTable, data_source: CsvDataSource) -> None:
    record_adapter = RecordIdAdapter(partial(_id, source=data_source.name))
    extract_references = RecordExtraction(data_source, record_adapter, single_record)
    for ref in extract_references():
        id_table.put(ref)


@contextmanager
def timer(start_message: str):
    logging.info(start_message)
    time_start = time.time()
    yield
    time_end = time.time()
    log.info("%s time elapsed: %.2f seconds", start_message, time_end - time_start)


@timer(start_message="train ditto")
def run_training():
    id_table = InMemoryIdTable()
    load_data_source(id_table, abt)
    load_data_source(id_table, buy)
    original_comparison_space_size = len(abt) * len(buy)
    comparison_space = create_comparison_space(
        id_table, gt, original_comparison_space_size
    )

    tokenizer = AutoTokenizer.from_pretrained(BERT_MODEL_NAME)
    ds = DittoDataset(
        comparison_space,
        id_table,
        gt,
        tokenizer,
        left_cols=("name", "description", "price"),
        right_cols=("name", "description", "manufacturer", "price"),
    )
    ditto = DittoModel(BERT_MODEL_NAME)
    train, xv, test = ds.split(3, 1, 1, 64)
    trainer = DittoTrainer(BERT_MODEL_NAME, Path(DATADIR).parent / "models", epochs=10)
    evaluator = DittoTrainingEvaluator(BERT_MODEL_NAME, xv, test)
    trainer.run_training(ditto, train, evaluator, True)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, stream=sys.stdout)
    with warnings.catch_warnings(action="ignore"):
        run_training()



import itertools
import random
import re
from collections.abc import Sequence, Generator
from typing import Callable

import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader, Subset
from transformers import PreTrainedTokenizerFast

from matchescu.matching.ml.ditto._ref_adapter import to_text
from matchescu.reference_store.comparison_space import BinaryComparisonSpace
from matchescu.reference_store.id_table import IdTable
from matchescu.typing import EntityReferenceIdentifier


def alphanumeric(token):
    return "".join([ch if ch.isalnum() else " " for ch in token])


class Augmenter(object):
    """Data augmentation operator.

    Support both span and attribute level augmentation operators.
    """

    def __init__(self):
        self._augmentations: dict[
            str, Callable[[Sequence, Sequence], tuple[Sequence, Sequence]]
        ] = {
            "del": self._del,
            "swap": self._swap,
            "drop_len": self._drop_len,
            "drop_sym": self._drop_symbols,
            "drop_same": self._drop_same,
            "drop_token": self._drop_token,
            "ins": self._insert,
            "append_col": self._append_col,
            "drop_col": self._drop_col,
        }

    def augment(self, tokens: Sequence[str], labels: Sequence[str], op="del"):
        augmentation = self._augmentations.get(op, lambda x, y: (x, y))
        return augmentation(labels, tokens)

    @staticmethod
    def _drop_col(labels, tokens):
        col_starts = [i for i in range(len(tokens)) if tokens[i] == "COL"]
        col_ends = [0] * len(col_starts)
        col_lens = [0] * len(col_starts)
        for i, pos in enumerate(col_starts):
            if i == len(col_starts) - 1:
                col_lens[i] = len(tokens) - pos
                col_ends[i] = len(tokens) - 1
            else:
                col_lens[i] = col_starts[i + 1] - pos
                col_ends[i] = col_starts[i + 1] - 1

            if tokens[col_ends[i]] == "[SEP]":
                col_ends[i] -= 1
                col_lens[i] -= 1
        candidates = [i for i, le in enumerate(col_lens) if le <= 8]
        if len(candidates) > 0:
            idx = random.choice(candidates)
            start, end = col_starts[idx], col_ends[idx]
            new_tokens = tokens[:start] + tokens[end + 1 :]
            new_labels = labels[:start] + labels[end + 1 :]
        else:
            new_tokens, new_labels = tokens, labels
        return new_labels, new_tokens

    @staticmethod
    def _append_col(labels, tokens):
        col_starts = [i for i in range(len(tokens)) if tokens[i] == "COL"]
        col_ends = [0] * len(col_starts)
        col_lens = [0] * len(col_starts)
        for i, pos in enumerate(col_starts):
            if i == len(col_starts) - 1:
                col_lens[i] = len(tokens) - pos
                col_ends[i] = len(tokens) - 1
            else:
                col_lens[i] = col_starts[i + 1] - pos
                col_ends[i] = col_starts[i + 1] - 1

            if tokens[col_ends[i]] == "[SEP]":
                col_ends[i] -= 1
                col_lens[i] -= 1
                break
        candidates = [i for i, le in enumerate(col_lens) if le > 0]
        if len(candidates) < 2:
            return tokens, labels
        else:
            idx1, idx2 = random.sample(candidates, k=2)
            start1, end1 = col_starts[idx1], col_ends[idx1]
            sub_tokens = tokens[start1 : end1 + 1]
            sub_labels = labels[start1 : end1 + 1]
            val_pos = 0
            for i, token in enumerate(sub_tokens):
                if token == "VAL":
                    val_pos = i + 1
                    break
            sub_tokens = sub_tokens[val_pos:]
            sub_labels = sub_labels[val_pos:]

            end2 = col_ends[idx2]
            new_tokens = []
            new_labels = []
            for i in range(len(tokens)):
                if start1 <= i <= end1:
                    continue
                new_tokens.append(tokens[i])
                new_labels.append(labels[i])
                if i == end2:
                    new_tokens += sub_tokens
                    new_labels += sub_labels
            return new_tokens, new_labels

    def _insert(self, labels, tokens):
        pos = self.sample_position(labels)
        symbol = random.choice("-*.,#&")
        new_tokens = tokens[:pos] + [symbol] + tokens[pos:]
        new_labels = labels[:pos] + ["O"] + labels[pos:]
        return new_tokens, new_labels

    @staticmethod
    def _drop_token(labels, tokens):
        new_tokens, new_labels = [], []
        for token, label in zip(tokens, labels):
            if label != "O" or random.randint(0, 4) != 0:
                new_tokens.append(token)
                new_labels.append(label)
        return new_labels, new_tokens

    @staticmethod
    def _drop_same(labels, tokens):
        left_token = set([])
        right_token = set([])
        left = True
        for token, label in zip(tokens, labels):
            if label == "O":
                token = token.lower()
                if left:
                    left_token.add(token)
                else:
                    right_token.add(token)
            if token == "[SEP]":
                left = False
        same = left_token & right_token
        targets = random.choices(list(same), k=1)
        new_tokens, new_labels = [], []
        for token, label in zip(tokens, labels):
            if token.lower() not in targets or label != "O":
                new_tokens.append(token)
                new_labels.append(label)
        return new_labels, new_tokens

    @staticmethod
    def _drop_symbols(labels, tokens):
        dropped_tokens = [alphanumeric(token) for token in tokens]
        new_tokens = []
        new_labels = []
        for token, d_token, label in zip(tokens, dropped_tokens, labels):
            if random.randint(0, 4) != 0 or label != "O":
                new_tokens.append(token)
                new_labels.append(label)
            else:
                if d_token != "":
                    new_tokens.append(d_token)
                    new_labels.append(label)
        return new_labels, new_tokens

    def _swap(self, labels, tokens):
        span_len = random.randint(2, 4)
        pos1, pos2 = self.sample_span(tokens, labels, span_len=span_len)
        if pos1 < 0:
            return tokens, labels
        sub_arr = tokens[pos1 : pos2 + 1]
        random.shuffle(sub_arr)
        new_tokens = tokens[:pos1] + sub_arr + tokens[pos2 + 1 :]
        new_labels = tokens[:pos1] + ["O"] * (pos2 - pos1 + 1) + labels[pos2 + 1 :]
        return new_tokens, new_labels

    def _del(self, labels, tokens):
        span_len = random.randint(1, 2)
        pos1, pos2 = self.sample_span(tokens, labels, span_len=span_len)
        if pos1 < 0:
            return tokens, labels
        new_tokens = tokens[:pos1] + tokens[pos2 + 1 :]
        new_labels = tokens[:pos1] + labels[pos2 + 1 :]
        return new_tokens, new_labels

    @staticmethod
    def _drop_len(labels, tokens):
        all_lens = [len(token) for token, label in zip(tokens, labels) if label == "O"]
        if len(all_lens) == 0:
            return tokens, labels
        target_lens = random.choices(all_lens, k=1)
        new_tokens = []
        new_labels = []

        for token, label in zip(tokens, labels):
            if label != "O" or len(token) not in target_lens:
                new_tokens.append(token)
                new_labels.append(label)
        return new_tokens, new_labels

    def augment_sent(self, text, op="all"):
        """Performs data augmentation on a classification example.

        Similar to augment(tokens, labels) but works for sentences
        or sentence-pairs.

        Args:
            text (str): the input sentence
            op (str, optional): a string encoding of the operator to be applied

        Returns:
            str: the augmented sentence
        """
        # 50% of chance of flipping
        if " [SEP] " in text and random.randint(0, 1) == 0:
            left, right = text.split(" [SEP] ")
            text = right + " [SEP] " + left

        # tokenize the sentence
        tokens = text.split(" ")

        # avoid the special tokens
        labels = []
        for token in tokens:
            if token in ["COL", "VAL"]:
                labels.append("HD")
            elif token in ["[CLS]", "[SEP]"]:
                labels.append("<SEP>")
            else:
                labels.append("O")

        if op == "all":
            # RandAugment: https://arxiv.org/pdf/1909.13719.pdf
            N = 3
            ops = ["del", "swap", "drop_col", "append_col"]
            for op in random.choices(ops, k=N):
                tokens, labels = self.augment(tokens, labels, op=op)
        else:
            tokens, labels = self.augment(tokens, labels, op=op)
        results = " ".join(tokens)
        return results

    @staticmethod
    def _process_index(
        idx: int, span_len: int, labels: Sequence[str]
    ) -> Generator[tuple[int, int]]:
        span_end = idx + span_len
        if span_end - 1 >= len(labels):
            return
        span_text = "".join(labels[idx:span_end])
        pattern = "^O{%d}$" % span_len
        if not re.match(pattern, span_text, re.S):
            return
        yield idx, span_end - 1

    def _enumerate_spans(
        self, tokens: Sequence[str], labels: Sequence[str], span_len: int = 3
    ) -> Generator[tuple[int, int]]:
        yield from (
            t
            for idx, _ in enumerate(tokens)
            for t in self._process_index(idx, span_len, labels)
        )

    def sample_span(
        self, tokens: Sequence[str], labels: Sequence[str], span_len: int = 3
    ) -> tuple[int, int]:
        candidates = list(self._enumerate_spans(tokens, labels, span_len))
        if len(candidates) == 0:
            return -1, -1
        return random.choice(candidates)

    @staticmethod
    def sample_position(labels):
        candidates = list(
            map(lambda x: x[0], filter(lambda x: x[1] == "O", enumerate(labels)))
        )
        return random.choice(candidates) if len(candidates) > 0 else -1


class DittoDataset(Dataset):
    """EM dataset"""

    def __init__(
        self,
        comparison_space: BinaryComparisonSpace,
        id_table: IdTable,
        ground_truth: set[tuple[EntityReferenceIdentifier, EntityReferenceIdentifier]],
        tokenizer: PreTrainedTokenizerFast,
        max_len=256,
        size=None,
        augmentations=None,
        left_cols: tuple | None = None,
        right_cols: tuple | None = None,
    ):
        self.__tokenizer = tokenizer
        id_pairs = list(itertools.islice(comparison_space, size))
        self.__pairs = list(tuple(map(id_table.get_all, id_pairs)))
        self.__labels = list(map(lambda pair: int(pair in ground_truth), id_pairs))
        self.__max_len = max_len
        self.__size = size
        self.__left_cols = left_cols
        self.__right_cols = right_cols
        self.da = augmentations
        if augmentations is not None:
            self.augmenter = Augmenter()
        else:
            self.augmenter = None

    def __len__(self):
        """Return the size of the dataset."""
        return len(self.__pairs)

    def __getitem__(self, idx):
        """Return a tokenized item of the dataset.

        Args:
            idx (int): the index of the item

        Returns:
            List of int: token ID's of the two entities
            List of int: token ID's of the two entities augmented (if da is set)
            int: the label of the pair (0: unmatch, 1: match)
        """
        left, right = self.__pairs[idx]
        left_text = to_text(left, self.__left_cols)
        right_text = to_text(right, self.__right_cols)
        x = self.__tokenizer.encode(
            text=left_text,
            text_pair=right_text,
            max_length=self.__max_len,
            truncation=True,
        )

        # augment if da is set
        if self.da is not None:
            augmented = self.augmenter.augment_sent(
                f"{left_text} [SEP] {right_text}", self.da
            )
            left_aug, right_aug = augmented.split(" [SEP] ")
            x_aug = self.__tokenizer.encode(
                text=left_aug,
                text_pair=right_aug,
                max_length=self.__max_len,
                truncation=True,
            )
            return x, x_aug, self.__labels[idx]
        else:
            return x, self.__labels[idx]

    @staticmethod
    def __pad(x: Sequence, total_length: int) -> torch.LongTensor:
        tensor_data = list(
            map(
                lambda vec: list(
                    itertools.chain(
                        vec, itertools.repeat(0, max(total_length - len(vec), 0))
                    )
                ),
                x,
            )
        )
        return torch.LongTensor(tensor_data)

    @staticmethod
    def __zero_padded(batch: list[tuple]) -> tuple[torch.LongTensor, ...]:
        if len(batch[0]) == 3:
            x1, x2, y = zip(*batch)

            n = max(map(len, x1 + x2))
            x1 = DittoDataset.__pad(x1, n)
            x2 = DittoDataset.__pad(x2, n)
            return x1, x2, torch.LongTensor(y)
        else:
            x, y = zip(*batch)
            n = max(map(len, x))
            x = DittoDataset.__pad(x, n)
            return x, torch.LongTensor(y)

    def split(
        self,
        train_proportion: float,
        xv_proportion: float,
        test_proportion: float,
        batch_size: int = 32,
    ) -> tuple[DataLoader, DataLoader, DataLoader]:
        proportion_total = train_proportion + xv_proportion + test_proportion
        train_proportion /= proportion_total
        xv_proportion /= proportion_total
        test_proportion = 1 - train_proportion - xv_proportion
        assert train_proportion + xv_proportion + test_proportion == 1
        total_size = len(self.__pairs)

        indices = np.arange(len(self.__pairs))
        np.random.shuffle(indices)

        train_size = int(train_proportion * total_size)
        xv_size = int(xv_proportion * total_size)

        # Step 5: Split the dataset using the shuffled indices
        train_indices = indices[:train_size]
        xv_indices = indices[train_size : train_size + xv_size]
        test_indices = indices[train_size + xv_size :]

        train_dataset = Subset(self, train_indices)
        xv_dataset = Subset(self, xv_indices)
        test_dataset = Subset(self, test_indices)
        return (
            DataLoader(
                train_dataset,
                batch_size=batch_size,
                shuffle=True,
                num_workers=0,
                collate_fn=DittoDataset.__zero_padded,
            ),
            DataLoader(
                xv_dataset,
                batch_size=batch_size * 16,
                shuffle=False,
                num_workers=0,
                collate_fn=DittoDataset.__zero_padded,
            ),
            DataLoader(
                test_dataset,
                batch_size=batch_size * 16,
                shuffle=False,
                num_workers=0,
                collate_fn=DittoDataset.__zero_padded,
            ),
        )



from collections.abc import Iterable
from os import PathLike, curdir
from pathlib import Path

import torch
from transformers import PreTrainedTokenizerFast

from matchescu.matching.ml.ditto._ditto_module import DittoModel
from matchescu.matching.ml.ditto._ref_adapter import to_text
from matchescu.typing import EntityReference


class DittoMatcher(object):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizerFast,
        max_len: int = 256,
        model_dir: str | PathLike | None = None,
        left_cols: Iterable[str] | None = None,
        right_cols: Iterable[str] | None = None,
    ) -> None:
        model_dir = model_dir or curdir
        self.__tokenizer = tokenizer
        self.__max_len = max_len
        self.__model_dir = Path(model_dir)
        self.__model = None
        self.__threshold = 0.5
        self.__left_cols = left_cols
        self.__right_cols = right_cols

    @property
    def non_match_threshold(self) -> float:
        return 1 - self.__threshold

    @property
    def match_threshold(self) -> float:
        return self.__threshold

    def load_pretrained(self, model_name: str) -> None:
        model_path = self.__model_dir / model_name / "model.pt"
        if not model_path.exists():
            raise FileNotFoundError(model_path)
        model_dict = torch.load(model_path)
        self.__threshold = float(model_dict["threshold"])
        self.__model = DittoModel(model_name)
        self.__model.load_state_dict(model_dict["model"])

    def __call__(self, left: EntityReference, right: EntityReference) -> float:
        with torch.no_grad():
            encoded_text = torch.LongTensor(
                self.__tokenizer.encode(
                    text=to_text(left, self.__left_cols),
                    text_pair=to_text(right, self.__right_cols),
                    max_length=self.__max_len,
                    truncation=True,
                )
            ).unsqueeze(0)
            return torch.sigmoid(self.__model(encoded_text)).item()



from logging import Logger, getLogger
from os import PathLike
from pathlib import Path
from typing import Any, cast

import torch
from torch.nn import BCEWithLogitsLoss
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from transformers import get_linear_schedule_with_warmup

from matchescu.matching.ml.ditto._ditto_dataset import DittoDataset
from matchescu.matching.ml.ditto._ditto_module import DittoModel
from matchescu.matching.ml.ditto._ditto_training_evaluator import DittoTrainingEvaluator


class DittoTrainer:
    def __init__(
        self, task_name: str, model_dir: str | PathLike | None = None, **kwargs: Any
    ) -> None:
        self._task = task_name
        self._model_dir = Path(model_dir) if model_dir else Path(__file__).parent
        self._tb_log_dir = self._model_dir / self._task / "tensorboard"
        self._log = cast(
            Logger, kwargs.get("logger", getLogger(self.__class__.__name__))
        )
        self._epochs = int(kwargs.get("epochs", 20))
        self._learning_rate = float(kwargs.get("learning_rate", 3e-5))
        self._frozen_layer_count = int(kwargs.get("frozen_layer_count", 0))

    def _add_text_summary(
        self, summary_writer: SummaryWriter, step_no: int, fmt: str, *args: Any
    ) -> None:
        text = fmt % args
        summary_writer.add_text(self._task, f"{self._task}: {text}", step_no)

    def __get_device(self):
        device = torch.device("cpu")
        if torch.cuda.is_available():
            device = torch.cuda.current_device()
        elif torch.backends.mps.is_available():
            device = torch.device("mps:0")
        else:
            if torch.backends.mps.is_built():
                self._log.info("MPS built, but not available.")
            else:
                self._log.info("Not Mac, nor CUDA.")
        return device

    def _train_one_epoch(
        self,
        epoch: int,
        device: torch.device,
        model: DittoModel,
        train_iter: DataLoader,
        optimizer: Optimizer,
        scheduler: LRScheduler,
        summary_writer: SummaryWriter,
    ):
        total_loss = 0.0
        batch_no = 0

        try:
            loss_fn = BCEWithLogitsLoss().to(device)
            model.to(device)
            model.train(True)
            batch_loss = 0.0
            for i, batch in enumerate(train_iter):
                device_batch = tuple(item.to(device) for item in batch)
                optimizer.zero_grad()

                if len(device_batch) == 2:
                    x, y = device_batch
                    prediction = model(x.to(device))
                else:
                    x1, x2, y = device_batch
                    prediction = model(x1.to(device), x2.to(device))

                loss = loss_fn(prediction, y.to(device).float())

                loss.backward()
                optimizer.step()
                scheduler.step()

                step_loss = loss.item()
                total_loss += step_loss
                batch_loss += step_loss
                batch_no = i + 1
                if batch_no % 10 == 0:
                    batch_loss = batch_loss / 10
                    fmt = f"{self._task}: batch {batch_no}: avg loss over last 10 batches=%.4f"
                    self._add_text_summary(summary_writer, batch_no, fmt, batch_loss)
                    self._log.info(fmt, batch_loss)
                del loss
        finally:
            model.train(False)

        avg_loss = total_loss / batch_no if batch_no > 0 else 0
        summary_writer.add_scalar(f"{self._task}_avg_loss", avg_loss, epoch)
        self._log.info("epoch %d: avg loss=%.4f", epoch, avg_loss)

    def run_training(
        self,
        model: DittoModel,
        training_data: DataLoader[DittoDataset],
        evaluator: DittoTrainingEvaluator = None,
        save_model: bool = False,
    ):
        device = self.__get_device()
        model = model.with_frozen_bert_layers(self._frozen_layer_count)

        optimizer = torch.optim.AdamW(model.parameters(), lr=self._learning_rate)
        total_batches = (
            len(cast(DittoDataset, training_data.dataset)) // training_data.batch_size
        )
        num_steps = total_batches * self._epochs
        scheduler = get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=0, num_training_steps=num_steps
        )

        summary_writer = SummaryWriter(log_dir=str(self._tb_log_dir.absolute()))
        best_xv_f1 = best_test_f1 = 0.0
        try:
            for epoch in range(1, self._epochs + 1):
                self._log.info("%s: epoch %d - train start", self._task, epoch)
                try:
                    self._train_one_epoch(
                        epoch,
                        device,
                        model,
                        training_data,
                        optimizer,
                        scheduler,
                        summary_writer,
                    )
                finally:
                    self._log.info("%s: epoch %d - train end", self._task, epoch)

                if evaluator is None or not save_model:
                    continue

                xv_f1, test_f1, threshold = evaluator(model)
                if xv_f1 > best_xv_f1:
                    self._log.info("found new best F1. saving checkpoint")
                    best_xv_f1 = xv_f1
                    best_test_f1 = test_f1

                    task_model_dir = self._model_dir / self._task
                    task_model_dir.mkdir(parents=True, exist_ok=True)
                    ckpt_path = task_model_dir / "model.pt"
                    ckpt = {
                        "model": model.state_dict(),
                        "optimizer": optimizer.state_dict(),
                        "scheduler": scheduler.state_dict(),
                        "epoch": epoch,
                        "threshold": threshold,
                    }
                    torch.save(ckpt, ckpt_path)

                scalars = {"f1": xv_f1, "t_f1": test_f1}
                summary_writer.add_scalars(self._task, scalars, epoch)
                self._log.info(
                    "xv_f1=%.4f, best_xv_f1=%.4f, test_f1=%.4f, best_test_f1=%.4f",
                    xv_f1,
                    best_xv_f1,
                    test_f1,
                    best_test_f1,
                )
        finally:
            summary_writer.flush()
            summary_writer.close()



import itertools
from typing import Hashable, Callable

import polars as pl

from matchescu.csg import BinaryComparisonSpaceGenerator
from matchescu.extraction import EntityReferenceExtraction
from matchescu.matching.entity_reference import (
    EntityReferenceComparisonConfig,
)
from matchescu.matching.ml.datasets._sampling import (
    AttributeComparison,
    PatternEncodedComparison,
)
from matchescu.reference_store.id_table._in_memory import InMemoryIdTable
from matchescu.typing import EntityReference, EntityReferenceIdentifier


class BlockDataSet:
    __TARGET_COL = "y"

    def __init__(
        self,
        csg: BinaryComparisonSpaceGenerator,
        id_table: InMemoryIdTable,
        left_id: Callable[[EntityReference], Hashable],
        right_id: Callable[[EntityReference], Hashable],
        ground_truth: set[tuple[Hashable, Hashable]],
    ) -> None:
        self.__csg = csg
        self.__id_table = id_table
        self.__lid = left_id
        self.__rid = right_id
        self.__true_matches = ground_truth
        self.__comparison_data = None
        self.__sample_factory = None

    @property
    def target_vector(self) -> pl.DataFrame:
        if self.__comparison_data is None:
            raise ValueError("comparison matrix was not computed")
        return self.__comparison_data[self.__TARGET_COL]

    @property
    def feature_matrix(self) -> pl.DataFrame:
        if self.__comparison_data is None:
            raise ValueError("comparison matrix was not computed")
        return self.__comparison_data.drop([self.__TARGET_COL])

    @staticmethod
    def __with_col_suffix(
        extract: EntityReferenceExtraction, suffix: str
    ) -> pl.DataFrame:
        df = pl.DataFrame(extract())
        return df.rename({key: f"{key}{suffix}" for key in df.columns})

    def attr_compare(self, config: EntityReferenceComparisonConfig) -> "BlockDataSet":
        self.__sample_factory = AttributeComparison(
            self.__true_matches,
            config,
            self.__lid,
            self.__rid,
            self.__TARGET_COL,
        )
        return self

    def pattern_encoded(
        self, config: EntityReferenceComparisonConfig, possible_outcomes: int = 2
    ) -> "BlockDataSet":
        self.__sample_factory = PatternEncodedComparison(
            self.__true_matches,
            config,
            self.__lid,
            self.__rid,
            self.__TARGET_COL,
            possible_outcomes,
        )
        return self

    def __concat_refs(self, *ref_ids: EntityReferenceIdentifier) -> tuple[tuple, int]:
        refs = list(map(self.__id_table.get, ref_ids))
        return tuple(v for r in refs for v in r), len(refs[0])

    def cross_sources(self) -> "BlockDataSet":
        data = [
            x[0]
            for x in itertools.starmap(
                self.__sample_factory,
                itertools.starmap(self.__concat_refs, self.__csg()),
            )
        ]
        self.__comparison_data = pl.DataFrame(data)
        return self



import itertools
import logging

import numpy as np
import torch
from sklearn import metrics
from torch.utils.data import DataLoader

from matchescu.matching.ml.ditto._ditto_dataset import DittoDataset
from matchescu.matching.ml.ditto._ditto_module import DittoModel


class DittoTrainingEvaluator:
    def __init__(
        self,
        task_name: str,
        xv_data: DataLoader[DittoDataset],
        test_data: DataLoader[DittoDataset],
    ) -> None:
        self._task = task_name
        self._xv_data = xv_data
        self._test_data = test_data
        self._log = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def __best_threshold(
        probabilities: np.ndarray, labels: np.ndarray
    ) -> tuple[float, float]:
        thresholds = np.arange(0.0, 1.0, 0.05)
        predictions = (probabilities[:, None] > thresholds).astype(int)
        f1_scores = np.fromiter(
            itertools.starmap(
                metrics.f1_score, zip(itertools.repeat(labels), predictions.T)
            ),
            dtype=np.float32,
        )
        best_idx = np.argmax(f1_scores)
        return float(f1_scores[best_idx]), float(thresholds[best_idx])

    @staticmethod
    @torch.no_grad()
    def _evaluate_model(
        model: DittoModel, data_loader: DataLoader, threshold: float | None = None
    ):
        model.eval()

        batch_results = map(lambda b: (torch.sigmoid(model(b[0])), b[1]), data_loader)
        all_probs, all_y = zip(*batch_results)
        all_probs = torch.cat(all_probs).detach().cpu().numpy()
        all_y = torch.cat(all_y).detach().cpu().numpy()

        if threshold is not None:
            pred = (all_probs > threshold).astype(int)
            f1 = metrics.f1_score(all_y, pred)
            return f1, threshold
        return DittoTrainingEvaluator.__best_threshold(all_probs, all_y)

    def __call__(self, model: DittoModel) -> tuple[float, float, float]:
        self._log.info("%s: evaluating on cross-validation set", self._task)
        xv_f1, best_xv_threshold = self._evaluate_model(model, self._xv_data)
        self._log.info(
            "%s: X validation F1=%.4f, best threshold=%.2f",
            self._task,
            xv_f1,
            best_xv_threshold,
        )
        test_f1, _ = self._evaluate_model(
            model, self._test_data, threshold=best_xv_threshold
        )
        self._log.info("%s: test F1=%.4f", self._task, test_f1)
        return xv_f1, test_f1, best_xv_threshold
